@@ -99,35 +99,20 @@ sub xyzzy_request_response {
 	$options{PATH_INFO} //= $path;
 	$options{QUERY_STRING} //= $query // '';
 
-	my $cfg = $handler
-		if UNIVERSAL::isa($handler, 'Xyzzy');
-
 	$handler = xyzzy_from_file($handler)
 		unless ref $handler;
 
-	$cfg = $handler
-		if UNIVERSAL::isa($handler, 'Xyzzy');
-
 	$handler = xyzzy_from_config(@$handler)
 		unless blessed $handler;
 
-	$cfg = $handler
-		if UNIVERSAL::isa($handler, 'Xyzzy');
-
 	$handler = $handler->handler
 		if UNIVERSAL::isa($handler, 'Xyzzy');
 
-	$cfg = $handler
-		if UNIVERSAL::isa($handler, 'Xyzzy');
-
-	$cfg = $handler->cfg
-		unless defined $cfg;
-
 	my ($inbuf, $outbuf, $errbuf) = ($options{REQUEST_BODY} // '', '', '');
 	open(my $in, '<', \$inbuf) or die;
 	open(my $out, '>', \$outbuf) or die;
 	open(my $err, '>', \$errbuf) or die;
-	my $req = new Xyzzy::Request::Root(cfg => $cfg, in => $in, out => $out, err => $err, env => {%$default_env, %options});
+	my $req = new Xyzzy::Request::Root(in => $in, out => $out, err => $err, env => {%$default_env, %options});
 	my $res = eval { $handler->handle($req) };
 	unless($res) {
 		my $err = $@;
@@ -144,7 +129,8 @@ sub xyzzy_request_response {
 sub xyzzy_request {
 	my $res = xyzzy_request_response(@_);
 	my $content = $res->content;
-	utf8::downgrade($content);
+	confess("Xyzzy::Response content must be binary! Convert logical characters explicitly using Encode::encode or Encode::encode_utf8")
+		if utf8::is_utf8($content);
 	return $content;
 }
 
@@ -152,7 +138,8 @@ sub xyzzy_request_as_cgi {
 	my $res = xyzzy_request_response(@_);
 	my $as_cgi = $res->as_cgi;
 	$as_cgi = join('', @$as_cgi);
-	utf8::downgrade($as_cgi);
+	confess("Xyzzy::Response content and headers must be binary! Convert logical characters explicitly using Encode::encode or Encode::encode_utf8")
+		if utf8::is_utf8($as_cgi);
 	return $as_cgi;
 }
 
