@@ -52,7 +52,7 @@ die "Usage: xyzzy <configfile>\n"
 
 use FCGI;
 use IO::Handle;
-use POSIX qw(pause setsid);
+use POSIX qw(setsid);
 
 use Xyzzy;
 use Xyzzy::Header;
@@ -76,6 +76,7 @@ if(my $numprocs = $ENV{XYZZY_FCGI_CHILDREN} // $ENV{PHP_FCGI_CHILDREN}) {
 	my $debug = $ENV{XYZZY_FCGI_DEBUG};
 	my $alarm;
 	my $sigalarm;
+	my $sigwinch;
 	local our $alarmdie;
 
 	setsid();
@@ -103,7 +104,8 @@ if(my $numprocs = $ENV{XYZZY_FCGI_CHILDREN} // $ENV{PHP_FCGI_CHILDREN}) {
 	}
 
 	eval {
-		local $SIG{HUP} = sub { $backoff = 0 };
+		local $SIG{WINCH} = sub { $backoff = 0; kill(POSIX::SIGTERM, keys %running) };
+		local $SIG{HUP} = sub { $backoff = 0; kill(POSIX::SIGHUP, keys %running) };
 		local $SIG{TERM} =
 		local $SIG{INT} = sub { my $sig = shift; die "SIG$sig received, exiting\n" };
 		local $SIG{ALRM} = sub { setitimer(0, 0, 0); $sigalarm = 1; die "Timeout\n" if $alarmdie };
@@ -141,7 +143,9 @@ if(my $numprocs = $ENV{XYZZY_FCGI_CHILDREN} // $ENV{PHP_FCGI_CHILDREN}) {
 					if(%running) {
 						$pid = waitpid(-1, 0);
 					} elsif($alarm) {
-						pause();
+						sleep;
+					} else {
+						sleep .1;
 					}
 				};
 				my $e = $@;
@@ -245,6 +249,7 @@ my $benchmark;
 my $numreqs;
 my $handler;
 my $hupped;
+my $termed;
 
 sub load_config {
 	my $cfg = new Xyzzy;
@@ -268,10 +273,20 @@ sub reload_config {
 	}
 }
 
+sub sighup {
+	$hupped = 1;
+}
+
+sub sigterm {
+	die "SIGTERM received, exiting\n" if our $termok;
+	$termed = 1;
+}
+
 eval {
 	load_config();
 
-	$SIG{HUP} = sub { $hupped = 1 };
+	$SIG{HUP} = \&sighup;
+	$SIG{TERM} = \&sigterm;
 
 	$0 = "$class (new)";
 
@@ -281,14 +296,21 @@ eval {
 		my $cpu_proc_init = clock_gettime(CLOCK_PROCESS_CPUTIME_ID);
 		my $wall_proc_init = clock_gettime(CLOCK_MONOTONIC) - $wall_proc_start;
 
-		printf STDERR "initial: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_proc_init, $cpu_proc_init, 100.0 * $cpu_proc_init / $wall_proc_init;
+		warn sprintf("initial: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_proc_init, $cpu_proc_init, 100.0 * $cpu_proc_init / $wall_proc_init);
 	}
 
 	for(;;) {
 		$SIG{__WARN__} = \&log_warnings;
-		reload_config() while $hupped;
-		$SIG{HUP} = sub { $hupped = 1 };
-		if($! = -$fcgi->Accept) {
+		$SIG{HUP} = \&sighup unless $SIG{HUP} == \&sighup;
+		$SIG{TERM} = \&sigterm unless $SIG{TERM} == \&sigterm;
+		$! = - do {
+			# try to prevent signal race conditions
+			local our $termok = 1;
+			die "SIGTERM received, exiting\n" if $termed;
+			reload_config() while $hupped;
+			$fcgi->Accept;
+		};
+		if($!) {
 			next if $!{EINTR};
 			die "FCGI: accept(): $!\n";
 		}
@@ -343,7 +365,7 @@ eval {
 			my $cpu_req_total = clock_gettime(CLOCK_PROCESS_CPUTIME_ID) - $cpu_req_start;
 			my $wall_req_total = clock_gettime(CLOCK_MONOTONIC) - $wall_req_start;
 
-			printf STDERR "request: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_req_total, $cpu_req_total, 100.0 * $cpu_req_total / $wall_req_total;
+			warn sprintf("request: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_req_total, $cpu_req_total, 100.0 * $cpu_req_total / $wall_req_total);
 		}
 
 		$0 = "$class (idle #$numreqs)";
@@ -356,7 +378,7 @@ if($benchmark) {
 	my $cpu_proc_total = clock_gettime(CLOCK_PROCESS_CPUTIME_ID);
 	my $wall_proc_total = clock_gettime(CLOCK_MONOTONIC) - $wall_proc_start;
 
-	printf STDERR "process: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_proc_total, $cpu_proc_total, 100.0 * $cpu_proc_total / $wall_proc_total;
+	warn sprintf("process: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_proc_total, $cpu_proc_total, 100.0 * $cpu_proc_total / $wall_proc_total);
 }
 
 if($err) {
